/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

// @(#) internal-src/lib/dacs/cell/common/dacs_internal.h v1.48 - 9/26/07 11:52:28 @(#)

#ifndef _DACS_INTERNAL_H_
#define _DACS_INTERNAL_H_

//
// Internal definitions common to SPE and PPE
//

#include <dacs.h>
#include <dacsi_q.h>

// Signature string 'DaCS' to verify correctness of the DaCS Control Block
#define DACS_CB_SIGNATURE   0x44614353

// Generic macro for aligning data on routine stacks.  
// Note: This is only needed as long as the aligned attribute does not work on
//       the stack.
#define ALIGNED_DATA(__name, __type, __align)                       \
    unsigned char __name##_data[(__align) + sizeof(__type)];        \
    __type *__name =                                                \
      (__type *)((unsigned long)&__name##_data[(__align)] & ~((__align)-1));

//
// PPE->SPE control block entry.  the PPE should set
// this up when it runs dacs_ppu_runtime_init().
// When the spe's are started, they should copy the control block array to
// Local Store.
// \note   The control block array is meant to hold \b static information only.
// \note   The pad declaration at the bottom is supposed to be a portable
//         way to pad the struct out to a multiple of of 16 bytes, to
//         avoid DMA crashes.  The anonymous union currently in use is not
//         as portable, but has the advanwide of actually compiling correctly.
//
typedef struct dacsi_cb
{
    union
    {
        struct
        {
        uint32_t          signature;            // 4 bytes: 'DaCS'
        de_id_t           de_id;                // 4 bytes: id for this DE
        DACS_DE_TYPE_T    type;                 // 4 bytes: type of DE
        dacs_process_id_t pid;                  // 8 bytes: process id, mostly 
                                                //     for the (static) PPE pid
        dacs_addr_64_t    jobQ[DACSI_JOBQ_CNT]; // 8*2 bytes: MS addr of 
                                                //     the message queues
        };
        unsigned char   pad[128];                  //< pad to 128 bytes
    };
} dacsi_cb_t __attribute__ ((aligned(16)));


//
// "handle" of the ae context.  On cell this maps to an speid, on systemX
// this should map to a pthread id.
//
typedef uint64_t de_handle_t;    // make it big enough to handle anything

//
// Process ID status
//
typedef enum DACSI_PID_STS
{
    DACSI_PID_INVALID=0x50,         //< 'P' pid status is invalid
    DACSI_PID_RUNNING,              //< 'Q' process is running
    DACSI_PID_STOPPED,              //< 'R' process has finished
}   DACSI_PID_STS_T;


//
// Descriptor of DE process termination information
// NOTE: This is the same as the libspe2 stopinfo struct and should be kept
//       the same.
//

typedef struct dacs_stop_info
{
    unsigned int stop_reason;
    union { 
        int exit_code; 
        int signal_code; 
        int runtime_error; 
        int runtime_exception; 
        int runtime_fatal; 
        int callback_error; 
        void *__reserved_ptr; 
        unsigned long long __reserved_u64; 
    } result; 
    int spu_status;
} dacs_stop_info_t;

//
// Descriptor of each process running on a de.
//
typedef struct  dacs_topo_pid
{
    dacs_process_id_t   pid;                // process/thread id
    void                *prog_handle;       // program handle
    DACSI_PID_STS_T     status;             // status
    unsigned int        flags;	            // thread flags
    unsigned int        entry;		    // thread PC
    dacs_stop_info_t    stopinfo;	    // context termination info
    unsigned int        nwaiters;           // number of waiters present
    dacs_addr_64_t      argv;               // 8 bytes: argv pointer
    dacs_addr_64_t      envv;               // 8 bytes: argv pointer

}   dacs_topo_pid_t;


//
// status of the de_id represented by this topology struct
typedef  enum   DACSI_TOPO_STS
{
    DACSI_TOPO_INVALID=0x47,        //< 'G' topo status is invalid
    DACSI_TOPO_FREE,                //< 'H' topo is free
    DACSI_TOPO_RESERVED,            //< 'I' topo is reserved
}   DACSI_TOPO_STS_T;

//
// Topology tree descriptor.   The topology is considered to be linked to
// the hardware structure, thus the pointer to the runtime control block.
//
// \note   this struct is padded to a multiple of 16 to avoid DMA crashes.
//         overkill?  not sure yet ;-)
//
typedef struct  dacs_topology
{
    union
    {
        struct
        {
            DACSI_TOPO_STS_T        reservation;        //< reserved flag
            uint32_t                my_index;        //< my index in the parent's children array :-0
            dacs_addr_64_t          ctx;             //< "handle" of the de context: speid, pthreadid, etc.
            dacsi_cb_t              *cb;               //< pointer to control block of the current DE.
            struct dacs_topology    *parent;            //< ptr to parent topology node
            uint32_t                num_peers;       //< number of peers
            struct dacs_topology    *peers;             //< ptr to array of peer topo's
            uint32_t                num_children;    //< number of children
            struct dacs_topology    *children;          //< ptr to array of children topo's
            uint16_t                nproc_supported;
            uint16_t                nproc_active;
            dacs_topo_pid_t         pids[MAX_PROCESSES_PER_DE];  //<    started, ended, failed?  One for each process.
            uint32_t dma_lock;                       // DMA synchronization
            dacs_addr_64_t mfc_area;
        };
        unsigned char pad[ 128 ];
    };  // end union
}   dacs_topology_t __attribute__ ((aligned(16)));


// Maximum # of spe wids allowed for PPE<==>SPE communications.  must
// be in the range 0-255
#define MAX_SPE_WIDS    32

// Maximum # of spe wids allowed for SYSTEMX<==>PPE communications.  must
// be in the range 256-511
#define MAX_PPE_WIDS    256

// Maximum individual DMA size.  All DMAs greater than this must be broken up
// into pieces of MAX size or smaller.
#define MAX_DMA_SIZE 0x4000

// associates errcodes with a specific string
typedef struct  dacsi_error_str
{
    DACS_ERR_T      err;        //< error number
    const char *    str;        //< string representing the error number
}   dacsi_error_str_t;


//  ===== internal global vars  ========================================
extern  de_id_t     dacspi_local_de_id;          // declared in dacs_*_runtime.c
extern  uint32_t    dacsi_runtime_initialized;  // declared in dacs_*_runtime.c

//  ----- inter-section calls (common functions, spe and ppe)   --------

// -----   runtime
extern  uint32_t           dacsi_is_initialized( void );

//  -----   topology
extern dacs_topology_t *dacsi_find_de( de_id_t de, dacs_topology_t * top );
extern dacs_topology_t *dacsi_get_topo( de_id_t de, DACS_ERR_T *errcode );
extern DACSI_TOPO_STS_T dacsi_check_reservation( de_id_t de );
extern int dacsi_check_children_reservations(dacs_topology_t *de_topo,
                                              DACSI_TOPO_STS_T checkval );
extern int dacsi_check_children_reservations(dacs_topology_t *de_topo,
                                              DACSI_TOPO_STS_T setval );

//  ----- internal debug routines --------------------------------------
extern void dacsi_dump_cb( const char *offset, dacsi_cb_t *cb );
extern void dacsi_dump_topo( const char * offset, dacs_topology_t * t );

#endif  //  _DACS_INTERNAL_H_
