/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
static char *ident = "@(#) internal-src/lib/dacs/cell/common/dacsi_q.c v1.19 - 9/24/07 13:33:06 @(#)";

#include <dacs.h>

#ifdef DACS_PPU
#include <dacs_ppe_internal.h>
#include <stdlib.h>
#endif

#ifdef DACS_SPU
#include <dacs_spe_internal.h>
#endif

#include <dacsi_q.h>
#include <dacs_debug.h>

//===========================================================================
// Message job Q specific utility routines
//===========================================================================
dacsi_jobQe_t *
dacsi_jobQe_peek(dacsi_jobQ_t *Q, uint32_t stream, DACSI_JOB_TYPE_T type)
{
    dacsi_jobQe_t *jobQe;
    if (dacsi_is_Q_active_empty((dacsi_Q_t*)Q)) {
        DPRINT("Qpeek active Q is empty\n");
        return NULL;
    }

    jobQe = ((dacsi_jobQe_t *)((unsigned long)Q + Q->Q.active.next));

    while (jobQe &&
           (jobQe->type != type ||
            (jobQe->stream != DACS_STREAM_ALL &&
            (stream != DACS_STREAM_ALL && jobQe->stream != stream)))) {

        jobQe = (jobQe->Q.next == Q->Q.active.off)
                ? 0 : (dacsi_jobQe_t*)DACSI_QE_NEXT(jobQe);
    }

    return jobQe;
}

dacsi_jobQe_t *
dacsi_jobQe_get(dacsi_jobQ_t *Q, uint32_t stream, DACSI_JOB_TYPE_T type)
{
    dacsi_jobQe_t *jobQe;

    jobQe = dacsi_jobQe_peek(Q, stream, type);

    if (jobQe)
        jobQe = (dacsi_jobQe_t *)dacsi_Q_remove((dacsi_Qe_t*)jobQe);

    return jobQe;
}

dacsi_jobQe_t *
dacsi_jobQe_alloc(dacsi_jobQ_t *Q)
{
    if (dacsi_is_Q_free_empty(Q))
        return 0;

    return (dacsi_jobQe_t *)dacsi_Q_getfree(Q);
}

static void
dacsi_jobQe_clear(dacsi_jobQe_t *job)
{
    job->stream = 0;
    job->initiator = 0;
    job->state = DACSI_JOB_ST_FREE;
    job->type =  DACSI_JOB_TYPE_INVALID;
    job->swap = 0;
    job->error = DACS_SUCCESS;
    job->len = 0;
    job->ms = 0;
}

dacsi_jobQe_t *
dacsi_jobQe_free(dacsi_jobQe_t *Qe)
{
    // We better have come in here with an entry that is not on a list
    ASSERT(!Qe->Q.next && !Qe->Q.prev);

    dacsi_jobQe_clear(Qe);

    dacsi_Q_putfree((dacsi_Qe_t *)Qe);

    ASSERT(Qe->Q.next && Qe->Q.prev);
    ASSERT(Qe->state == DACSI_JOB_ST_FREE);

    return NULL;
}

dacsi_jobQe_t *
dacsi_jobQe_put(dacsi_jobQe_t *Qe)
{
    return (dacsi_jobQe_t *)dacsi_Q_putactive((dacsi_Qe_t *)Qe);
}


//===========================================================================
// Wid element Q specific utility routines
//===========================================================================

DACS_ERR_T
dacsi_we_get(dacs_wid_t wid, de_id_t de, dacs_process_id_t pid, 
             dacsi_wid_elem_t **we, dacsi_waitQe_t **waitQe, dacs_tag_t *hwtag)
{
    DACS_ERR_T err = DACS_SUCCESS;

    DACSI_LOCK_WID_QUEUE(wid);

    // Search for a matching DE/PID.  If one is found then we'll add to it
    // rather than creating a new one.
    *we = dacsi_waitq[wid].head;
    while (*we) {
        if ((*we)->de == de && (*we)->pid == pid) {
            if (hwtag) {
                // If we have not yet allocated a hwtag to the we, then do so
                // now.
                if ((*we)->hwtag == -1U)
                    (*we)->hwtag = DACSI_TAG_RESERVE();

                *hwtag = (*we)->hwtag;
            }
            break;
        }
        *we = (*we)->next;
    }

    // Did not find one, so now we have to create one and initialize it.
    if (!*we) {
        dacsi_waitQ_t *Q;

        // Allocate storage for the wait element
        // We do this here for two reasons:
        //      1) We want to make sure it succeeds before starting the DMA
        //      2) The create routine is common between PPU and SPU, yet the
        //      allocation methods are different.
        // TODO: This may perform better with static memory, at least for the
        //       Q head.
        *we = DACSI_WE_ALLOC();
        if (!*we) {
            DPRINT("Failed to allocate a wait element.\n");
            err = DACS_ERR_NO_RESOURCE;
        }

        if (!err) {
            (*we)->request = ((unsigned long)(*we)) + sizeof(dacsi_wid_elem_t);
            Q = (dacsi_waitQ_t *)DACSI_64_TO_32((*we)->request);
            Q->first = Q->last = 0;
            (*we)->de = de;
            (*we)->pid = pid;

            // If the caller needs a hwtag, then reserve one for the we now.
            // Otherwise, set it to -1, indicating one has not been allocated.
            if (hwtag)
                *hwtag = (*we)->hwtag = DACSI_TAG_RESERVE();
            else
                (*we)->hwtag = -1;
        }
    }

    // Ok, we located/allocated a WID element up above.  
    // Now, allocate a wait queue element for this particular job.
    if (!err) {
        *waitQe = DACSI_WAITQE_ALLOC();
        if (!*waitQe) {
            dacsi_we_free(&dacsi_waitq[wid], *we);
            err = DACS_ERR_NO_RESOURCE;
        }
    }

    DACSI_UNLOCK_WID_QUEUE(wid);

    return err;
}

// Must be called under the WID queue lock
DACS_ERR_T
dacsi_we_free(dacsi_wait_queue_t *waitQ, dacsi_wid_elem_t *we)
{
    dacsi_wid_deq_head(waitQ);
    if (we->hwtag != -1U)
        DACSI_TAG_RELEASE(we->hwtag);
    DACSI_WE_FREE(we);

    return DACS_SUCCESS;
}


//===========================================================================
// Wait Q element specific utility routines
//===========================================================================

DACS_ERR_T
dacsi_waitQe_enq(dacs_wid_t wid, dacsi_wid_elem_t *we, dacsi_waitQe_t *waitQe,
                 DACSI_JOB_TYPE_T type, dacs_addr_64_t job)
{
    dacsi_waitQ_t *Q = (dacsi_waitQ_t *)DACSI_64_TO_32(we->request);

    DACSI_LOCK_WID_QUEUE(wid);

    waitQe->next = NULL;
    waitQe->job = job;
    waitQe->type = type;

    // If we have a Q entry, then a wait element has already been
    // queued.  If not, then we need to simply add one now.
    if (Q->first == 0 && Q->last == 0) {
        Q->first = Q->last = waitQe;

        // Add the wait element for this job to the wait queue.
        dacsi_wid_enq_tail(wid, we);
    }

    // An entry has already been queued ... add another at the tail
    else {
        Q->last = Q->last->next = waitQe;
    }

    DACSI_UNLOCK_WID_QUEUE(wid);

    return DACS_SUCCESS;
}

// Must be called under the WID queue lock
DACSI_JOB_TYPE_T 
dacsi_waitQe_get (dacsi_wid_elem_t *we, dacs_addr_64_t *job) 
{
    dacsi_waitQ_t *Q = (dacsi_waitQ_t *)DACSI_64_TO_32(we->request);
    dacsi_waitQe_t *next;

    // Nothing to do if the Q is empty
    // This is not an error to call in this scenario because someone could be
    // calling this routine in a loop as they walk the jobs.
    if (!Q->first && !Q->last)
        return DACSI_JOB_TYPE_INVALID;

    *job = Q->first->job;

    return Q->first->type;
}

// Must be called under the WID queue lock
DACS_ERR_T
dacsi_waitQe_free (dacsi_wid_elem_t *we) 
{
    dacsi_waitQ_t *Q = (dacsi_waitQ_t *)DACSI_64_TO_32(we->request);
    dacsi_waitQe_t *next;

    next = Q->first->next;
    DACSI_WAITQE_FREE(Q->first);
    Q->first = next;

    // A zero first element means we are done
    if (!Q->first) {
        Q->last = NULL;
    }

    return DACS_SUCCESS;
}

//===========================================================================
// Q Debug utility routines
//===========================================================================

#ifdef NEED_DEBUG
void
dacsi_Q_print_info(dacsi_Q_t *Q)
{
    printf("Q = %p\n", Q);
    printf("Q->active = %p\n", &Q->active);
    printf("Q->free = %p\n", &Q->free);
    printf("Q->active.next = 0x%x\n", Q->active.next);
    printf("Q->active.prev = 0x%x\n", Q->active.prev);
    printf("Q->active.off = 0x%x\n", Q->active.off);
    printf("Q->free.next = 0x%x\n", Q->free.next);
    printf("Q->free.prev = 0x%x\n", Q->free.prev);
    printf("Q->free.off = 0x%x\n", Q->free.off);
}

void
dacsi_Q_printQ(dacsi_Q_t *Q)
{
    dacsi_Qe_t *Qe;
    int cnt;

    Qe = (dacsi_Qe_t *)((unsigned long)Q + Q->free.next);
    if (Qe != &Q->free) {
        printf("\nFREE\n");
        printf("     prev       Qe              next\n");
        cnt = 1;
        while (Qe != &Q->free) {
            printf("%d: 0x%04x <- 0x%04x -> 0x%04x\n", 
                   cnt++, Qe->prev, Qe->off, Qe->next);
            Qe = (dacsi_Qe_t *)((unsigned long)Q + Qe->next);
        }
    }

    Qe = (dacsi_Qe_t *)((unsigned long)Q + Q->active.next);
    if (Qe != &Q->active) {
        printf("\nACTIVE\n");
        printf("     prev       Qe              next\n");
        cnt = 1;
        while (Qe != &Q->active) {
            printf("%d: 0x%04x <- 0x%04x -> 0x%04x\n", 
                   cnt++, Qe->prev, Qe->off, Qe->next);
            Qe = (dacsi_Qe_t *)((unsigned long)Q + Qe->next);
        }
    }

    printf("\n");
}



void
dacsi_jobQ_printjob(dacsi_jobQe_t *job)
{
    printf("\nJOB %p\n", job);
    printf("\tnext = 0x%x\n", job->Q.next);
    printf("\tprev = 0x%x\n", job->Q.prev);
    printf("\tstate = %d\n", job->state);
    printf("\tjobid = %d\n", job->jobid);
    printf("\tinitiator = %d\n", job->initiator);
    printf("\ttype = %d\n", job->type);
    printf("\tswap = %d\n", job->swap);
    printf("\tQtype = %d\n", job->Qtype);
    printf("\tQidx = %d\n", job->Qidx);
    printf("\terror = %d\n", job->error);
    printf("\tlen = %d\n", job->len);
    printf("\tbuffer = 0x%x%x\n", job->unused, job->ls);
}

#endif

