/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

/*
 * dacs_runtime.c - defines functions for initializing / deinitializing DaCS processes
 */

/*--------------------------------------------------------------------*/
/*  Includes                                                          */
/*--------------------------------------------------------------------*/

#include <dacs.h>
#include <dacs_common.h>
#include <dacs_error_internal.h> // For init/exit functions
#include <dacsi_wids.h>
#include <dacsi_dma.h>

#if defined(DACS_HYBRID) || defined(DACS_PPU)
#include <dacs_hybrid_runtime.h>

#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#endif

#ifdef DACS_PPU
#include <dacs_ppu_runtime.h>
#endif

#ifdef DACS_SPU
#include <dacs_spu_runtime.h>
#endif

TRACE_TIMERS_STRUCTURE_DEFINE;
TRACE_COUNTERS_STRUCTURE_DEFINE;

long dacsi_numa_node = -1;

DACS_ERR_T
dacs_runtime_init(void *arg1, void *arg2)
{
    DACS_ERR_T rc = DACS_SUCCESS;

    TRACE_INIT();

    TRACE_INTERVAL_TOKEN_ARGUMENT(token);
    TRACE_POINT_ENTRY(_DACS_RUNTIME_INIT,token,1,arg1,arg2);
    TRACE_TIMERS_STRUCTURE_INIT;
    TRACE_COUNTERS_STRUCTURE_INIT;

    /* the arguments are currently
     * unused and not supported
     */
    if (arg1 || arg2) {
        rc = DACS_ERR_INVALID_ADDR;
        goto err_out;
    }


#ifndef DACS_SPU
    /* start error handling thread */
    /* moving it so all threads get sigblock for sigterm */
    rc = dacsi_error_init();
    if (rc != DACS_SUCCESS) {
        dacsi_error_exit();
        goto err_out;   
    }
#endif

#if    defined(DACS_HYBRID) && !defined(DACS_PPU)
    rc = dacs_hybrid_runtime_init(arg1, arg2, &dacsi_local_de_id, &dacsi_local_pid);
#elif  defined(DACS_HYBRID) && defined(DACS_PPU)
    /*Both are called on the PPU.  Hybrid passes the de_id_t to ppu*/
    /*What happens if an error is returned on first function.  Should it skip next and just return?.  I think we must*/
    rc = dacs_hybrid_runtime_init(arg1, arg2, &dacsi_local_de_id, &dacsi_local_pid);

    if(rc == DACS_SUCCESS) {
        char * numa_id_str = getenv("DACS_HYBRID_INTERNAL_NUMA_NODE");
        if (numa_id_str != NULL) {
            char *endptr = NULL;
            unsigned long numa_id = strtol(numa_id_str,&endptr,0);

            if ((*endptr != 0) ||
                    (((numa_id == LONG_MAX) || (numa_id == LONG_MIN)) && (errno == ERANGE)))  {
                // Value was invalid - too big, not a number, etc...
                rc = DACS_ERR_INTERNAL;
            } else {
                // Value was converted, use it
                dacsi_numa_node = numa_id;
            }
        } else {
            // Value must be set
            rc = DACS_ERR_INTERNAL;
        }

        if (rc == DACS_SUCCESS) {
            // pass in the ppu's de
            rc = dacs_ppu_runtime_init(arg1, arg2, dacsi_local_de_id);
        }
    }
#elif  !defined(DACS_HYBRID) && defined(DACS_PPU)
    dacsi_local_de_id  = 0x01010000; //OPT = 01 BLADE = 01 
    dacsi_local_pid    = getpid();
    rc = dacs_ppu_runtime_init(arg1, arg2, dacsi_local_de_id);
#elif  defined(DACS_SPU)
    rc = dacs_spu_runtime_init(arg1, arg2);
#else
#error "Invalid platform or no platform specified"
#endif

#ifndef DACS_SPU
    /* error handling thread */
    if (rc != DACS_SUCCESS) {
        dacsi_error_exit();
    }
    else {
        dacsi_proc_sync_init();
        dacsi_wids_init();
        dacsi_remote_mem_init();
    }
#endif

    if (rc == DACS_SUCCESS) {
        dacsi_initialized = 1;
    }

err_out:
    TRACE_POINT_EXIT(_DACS_RUNTIME_INIT,token,1,rc);

    return rc;
}

/*--------------------------------------------------------------------*/
DACS_ERR_T dacs_runtime_exit()
{
    DACS_ERR_T rc = DACS_SUCCESS;

#ifdef DACS_ERROR_CHECKING
    if(!dacsi_is_init())                       return DACS_ERR_NOT_INITIALIZED;
#endif


    TRACE_INTERVAL_TOKEN_ARGUMENT(token);
    TRACE_POINT_ENTRY(_DACS_RUNTIME_EXIT,token,1,0);

    TRACE_COUNTERS_TRACE_POINT;
    TRACE_TIMERS_TRACE_POINT;

#if    defined(DACS_HYBRID) && !defined(DACS_PPU)
    rc = dacs_hybrid_runtime_exit();
#elif  defined(DACS_HYBRID) && defined(DACS_PPU)
    /*Both are called on the PPU. Both have to free resouce.  Does the order matter? */
    /*What happens if an error is returned on first function.  Should it skip next and just return?*/
    rc = dacs_ppu_runtime_exit();
    if(rc == DACS_SUCCESS) {
        rc = dacs_hybrid_runtime_exit();
    }
#elif  !defined(DACS_HYBRID) && defined(DACS_PPU)
    rc = dacs_ppu_runtime_exit();
#elif  defined(DACS_SPU)
    rc = dacs_spu_runtime_exit();
#else
#error "Invalid platform or no platform specified"
#endif

#ifndef DACS_SPU
    /* shut down error handling thread */
    if (rc == DACS_SUCCESS) {
        dacsi_proc_sync_exit();
        dacsi_wids_destroy();
        dacsi_remote_mem_exit();
        dacsi_error_exit();
    }
#endif

    if (rc == DACS_SUCCESS) {
        dacsi_initialized = 0;
    }

    TRACE_POINT_EXIT(_DACS_RUNTIME_EXIT,token,1,rc);

    return rc;
}
