/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

// @(#) internal-src/lib/dacs/common/dacsi_wids.h v1.8 - 9/24/07 13:42:05 @(#)

/*
 * dacs_wids.c - defines functions for reserving DaCS Wait IDs (wids)
 */

#ifndef DACSI_WID_H
#define DACSI_WID_H

/*--------------------------------------------------------------------*/
/*  Includes                                                          */
/*--------------------------------------------------------------------*/

#include <dacs.h>
#include <stdint.h>

#if defined(DACS_HYBRID) || defined(DACS_PPU)
#include <pthread.h>
#endif

/*--------------------------------------------------------------------*/
/*  Constants                                                         */
/*--------------------------------------------------------------------*/

enum dacsi_wid_elem_type
{
    DACSI_SEND      = 0x0001,
    DACSI_RECV      = 0x0002,
    DACSI_PUT       = 0x0004,
    DACSI_GET       = 0x0008,
    DACSI_PUTL      = 0x0010,
    DACSI_GETL      = 0x0020,

    DACSI_FENCE     = 0x1000,
    DACSI_BARRIER   = 0x2000
};


#if defined(DACS_HYBRID) || defined(DACS_PPU)
#define DACSI_WID_MAX 256
#elif defined(DACS_SPU)
#define DACSI_WID_MAX 32
#endif


/*--------------------------------------------------------------------*/
/*  User Types                                                        */
/*--------------------------------------------------------------------*/

typedef struct dacsi_wid_element
{
    struct dacsi_wid_element   *next;
    de_id_t                     de;
    dacs_process_id_t           pid;
    uint32_t                    type;
    uint64_t                    request;
    uint32_t                    hwtag;
} dacsi_wid_elem_t;

typedef struct dacsi_wait_queue
{
#if defined(DACS_HYBRID) || defined(DACS_PPU)
    pthread_mutex_t lock;
#endif
    uint32_t        flags;
#define DACSI_WID_RESERVED 0x80000000

    dacsi_wid_elem_t *head;
    dacsi_wid_elem_t *tail;
} dacsi_wait_queue_t;

/*--------------------------------------------------------------------*/
/*  Macros                                                            */
/*--------------------------------------------------------------------*/
#if defined(DACS_HYBRID) || defined(DACS_PPU)

#define DACSI_WID_QUEUES_LOCK_INIT() pthread_mutex_init(&dacsi_waitq_lock,NULL)
#define DACSI_LOCK_WID_QUEUES() pthread_mutex_lock(&dacsi_waitq_lock)
#define DACSI_UNLOCK_WID_QUEUES() pthread_mutex_unlock(&dacsi_waitq_lock)

#define DACSI_WID_QUEUE_LOCK_INIT(wid) pthread_mutex_init(&dacsi_waitq[(wid)].lock,NULL)
#define DACSI_LOCK_WID_QUEUE(wid) pthread_mutex_lock(&dacsi_waitq[(wid)].lock)
#define DACSI_TRYLOCK_WID_QUEUE(wid) pthread_mutex_trylock(&dacsi_waitq[(wid)].lock)
#define DACSI_UNLOCK_WID_QUEUE(wid) pthread_mutex_unlock(&dacsi_waitq[(wid)].lock)

#elif defined(DACS_SPU)

#define DACSI_WID_QUEUES_LOCK_INIT()
#define DACSI_LOCK_WID_QUEUES() 
#define DACSI_UNLOCK_WID_QUEUES()

#define DACSI_WID_QUEUE_LOCK_INIT(wid)
#define DACSI_LOCK_WID_QUEUE(wid)
#define DACSI_TRYLOCK_WID_QUEUE(wid) 0
#define DACSI_UNLOCK_WID_QUEUE(wid)

#else

#error "Invalid platform or no platform specified"

#endif

// We don't lock on this check so that it can be taken under lock.  Plus for
// lightweight debug checks a lock is probably not necessary.
#define DACSI_IS_WID_RESERVED(wid)                             \
        ((dacsi_waitq[wid].flags & DACSI_WID_RESERVED) == DACSI_WID_RESERVED)

/*--------------------------------------------------------------------*/
/*  Global Variables                                                  */
/*--------------------------------------------------------------------*/

#ifdef DACS_WIDS_C
#define GLOBAL
#else
#define GLOBAL extern
#endif

/**
 * Global lock for the all of the wait queues. This is used during
 * reserveation and releasing of wait identifiers to maintain the queue
 * array state
 */
#if defined(DACS_HYBRID) || defined(DACS_PPU)
GLOBAL pthread_mutex_t dacsi_waitq_lock;
#endif

/**
 * Array of wait queues. A wait identifier is an index into this array
 */
GLOBAL struct dacsi_wait_queue dacsi_waitq[DACSI_WID_MAX];

/*--------------------------------------------------------------------*/
/*  Function Prototypes                                               */
/*--------------------------------------------------------------------*/

/**
 * This function must be called before WIDs are used to initialize the
 * data structures associated with WIDs.
 */
void dacsi_wids_init();

/**
 * This function must be called before DaCS terminates to cleanup any
 * data structures associated with WIDs. It is assumed that all
 * outstanding requests at DaCS termination are cleand up prior to calling
 * this function
 */
void dacsi_wids_destroy();

/**
 * Lock the Queue associated with the specified wid. This allows the
 * client to lock the queue so no changes occur during its use. The user
 * would lock the queue so no additional operations are added/deleted or
 * the wid is not released. This lock should only be held for short
 * durations.
 *
 * @param wid [in] the wait identifier to queue the operation on
 *
 * @return validation status of tag
 * <ul>
 * <li>DACS_ERR_NO_ERROR - operation completed successfully
 * <li>DACS_ERR_WID_NOT_RESERVED - WID is no longer reserved for
 * operations
 * <li>DACS_ERR_INVALID_WID - WID has an out of range value
 * </ul>
 *
 * @see #dacs_wid_unlock
 */
DACS_ERR_T dacsi_wid_lock(dacs_wid_t wid);

/**
 * Unlock the queue associated with the specified WID. No error checking
 * is done on the wid, it is assumed that dacsi_wid_lock was called with
 * the same wid to ensure that the wid was valid.
 *
 * @pre #dacsi_wid_lock was called
 * @post wid queue is unlocked
 *
 * @param wid [in] the wait identifier to queue the operation on
 *
 * @return DACS_ERR_NO_ERROR on success or DACS_ERR_GENERIC if the
 * unlock failed
 *
 */
DACS_ERR_T dacsi_wid_unlock(dacs_wid_t wid);

/**
 * Add a new element to a wait queue. The element queued in FIFO order
 * on the wait queue
 *
 * @pre the WID queue is already locked
 * @post the operation is enqueued
 *
 * @param wid [in] the wait identifier to queue the operation on
 * @param elem [in] a pointer to the element to add
 *
 */
void dacsi_wid_enq_tail(dacs_wid_t wid, dacsi_wid_elem_t *elem);

/**
 * Remove the head element from the wait queue. It is assumed that the
 * wait queue is locked prior to calling this function
 *
 * @pre the WID queue is already locked
 * @post the operation at the head of the queue is dequeued
 *
 * @param waitq [in] the address of the wait queue to manipulate
 *
 * @return the head element or NULL if the queue is empty
 */
dacsi_wid_elem_t * dacsi_wid_deq_head(struct dacsi_wait_queue * waitq);

#undef GLOBAL

#endif
